from nose.tools import eq_

from ..apply import apply
from ..tokenizers import text_split


def diff_and_replay(diff, tokenizer=None):
    a = """
    This sentence is going to get copied. This sentence is going to go away.
    
    ASDSJDNA  asas random words.
    
    This is another sentence.
    """
    
    b = """
    This sentence is going to get copied.  Wha... a new thing appeared!
    
    ASDSJDNA  asas random words.
    
    This is another sentence. This sentence is going to get copied.
    """
    a_tokens = list(text_split.tokenize(a))
    b_tokens = list(text_split.tokenize(b))
    operations = list(diff(a_tokens, b_tokens))
    
    for op in operations:
        if op.name == "equal":
            print("equal: " + str(a_tokens[op.a1:op.a2]))
        elif op.name == "delete":
            print("delete: " + str(a_tokens[op.a1:op.a2]))
        elif op.name == "insert":
            print("insert: " + str(b_tokens[op.b1:op.b2]))
    
    replay_b = [str(t) for t in apply(operations, a_tokens, b_tokens)]
    eq_(b_tokens, replay_b)



from .apply import apply
from .operations import Operation, Insert, Delete, Equal
from .algorithms.diff_engine import DiffEngine
from .algorithms import segment_matcher, SegmentMatcher
from .algorithms import sequence_matcher, SequenceMatcher
from .tokenizers import Tokenizer, RegexTokenizer, text_split, wikitext_split
from .segmenters import Segmenter, Segment, MatchableSegment

__version__ = "0.2.5"



import yamlconf
"""
"""

class DiffEngine:
    """
    Constructs a diff engine.
    """

    class Processor:
        """
        Constructs a new diff processor for processing many versions
        of a single text.
        """

        def process(text):
            raise NotImplementedError()


    def processor():
        """
        Configures and returns a new :class:`deltas.algorithms.Engine.Processor`
        """
        raise NotImplementedError()

    @classmethod
    def from_config(cls, config, name, section_key="diff_engines"):
        """
        Constructs a :class:`deltas.algorithms.Engine` from a configuration
        doc.
        """
        section = config[section_key][name]
        if 'module' in section:
            return yamlconf.import_module(section['module'])
        else:
            Engine = yamlconf.import_module(section['class'])
            return Engine.from_config(config, name, section_key=section_key)



import re

from ..util import LookAhead
from .segmenter import Segmenter
from .segments import MatchableSegment, Segment

WHITESPACE = set(["whitespace", "break"])
PARAGRAPH_END = set(["break"])
SENTENCE_END = set(["period", "epoint", "qmark"])
MIN_SENTENCE = 3

class ParagraphsSentencesAndWhitespace(Segmenter):
    """
    Constructs a paragraphs, sentences and whitespace segmenter.  This segmenter
    is intended to be used in western languages where sentences and paragraphs
    are meaningful segments of text content.

    Tree structure:

    * whitespace : :class:`deltas.segmenters.Segment`
    * paragraph : :class:`deltas.segmenters.MatchableSegment`
     * sentence : :class:`deltas.segmenters.MatchableSegment`
     * whitespace : :class:`deltas.segmenters.Segment`

     :Example:
         >>> from deltas import ParagraphsSentencesAndWhitespace, text_split
         >>> from deltas.segmenters import print_tree
         >>>
         >>> a = text_split.tokenize("This comes first.  This comes second.")
         >>>
         >>> segmenter = ParagraphsSentencesAndWhitespace()
         >>> segments = segmenter.segment(a)
         >>>
         >>> print_tree(segments)
         MatchableSegment: 'This comes first.  This comes second.'
         	MatchableSegment: 'This comes first.'
         	Segment: '  '
         	MatchableSegment: 'This comes second.'

    :Parameters:
        whitespace : `set`(`str`)
            A set of token types that represent whitespace.
        paragraph_end : `set`(`str`)
            A set of token types that represent the end of a pragraph.
        sentence_end : `set`(`str`)
            A set of tokens types that represent the end of a sentence.
        min_sentence : int
            The minimum non-whitespace tokens that a sentence must contain
            before a sentence_end will be entertained.
    """
    def __init__(self, *, whitespace=None,
                          paragraph_end=None,
                          sentence_end=None,
                          min_sentence=None):

        self.whitespace = set(whitespace or WHITESPACE)
        self.paragraph_end = set(paragraph_end or PARAGRAPH_END)
        self.sentence_end = set(sentence_end or SENTENCE_END)
        self.min_sentence = int(min_sentence or MIN_SENTENCE)

    def segment(self, tokens):
        """
        Segments a sequence of tokens into a sequence of segments.

        :Parameters:
            tokens : `list`(`deltas.tokenizers.Token`)
        """
        look_ahead = LookAhead(tokens)

        segments = Segment()

        while not look_ahead.empty():

            if look_ahead.peek().type not in self.whitespace: # Paragraph!
                paragraph = MatchableSegment()

                while not look_ahead.empty() and \
                      look_ahead.peek().type not in self.paragraph_end:

                    if look_ahead.peek().type not in self.whitespace: #Sentence!
                        sentence = MatchableSegment([next(look_ahead)])

                        while not look_ahead.empty() and \
                              look_ahead.peek().type not in self.paragraph_end:

                            sentence.append(next(look_ahead))

                            if sentence[-1].type in self.sentence_end:
                                non_whitespace = sum(s.type not in self.whitespace for s in sentence)
                                if non_whitespace >= self.min_sentence:
                                    break


                        paragraph.append(sentence)

                    else: # look_ahead.peek().type in self.whitespace
                        whitespace = Segment([next(look_ahead)])
                        paragraph.append(whitespace)

                segments.append(paragraph)
            else: # look_ahead.peek().type in self.whitespace
                whitespace = Segment([next(look_ahead)])
                segments.append(whitespace)


        return segments

    @classmethod
    def from_config(cls, config, name, section_key="segmenters"):
        section = config[section_key][name]
        return cls(
            whitespace=section.get('whitespace'),
            paragraph_end=section.get('paragraph_end'),
            sentence_end=section.get('sentence_end'),
            min_sentence=section.get('min_sentence')
        )



from nose.tools import eq_

from ...tokenizers import Token
from ..segments import MatchableSegment, Segment


def test_matchable_segment():

    words = ["foo", "bar", "baz"]

    ms = MatchableSegment([Token.construct(c, i) for i, c in enumerate(words)])
    eq_(ms.start, 0)
    eq_(ms.end, len(words))
    hash(ms)
    eq_(ms.match, None)
    ms.match = "derp"
    eq_(ms.match, "derp")

    d = {}
    d[ms] = ms

    ms2 = MatchableSegment([Token.construct(c, i) for i, c in enumerate(words)])
    assert ms2 in d

def test_segment():

    words = ["foo", "bar", "baz"]

    ms = Segment([Token.construct(c, i) for i, c in enumerate(words)])
    eq_(ms.start, 0)
    eq_(ms.end, len(words))

def test_equality():
    eq_(MatchableSegment([Token.construct("zero", 0), Token.construct("one", 1)]),
        MatchableSegment([Token.construct("zero", 2), Token.construct("one", 4)]))
    eq_(MatchableSegment([
            MatchableSegment([Token.construct("zero", 0), Token.construct("one", 1)]),
            MatchableSegment([Token.construct("two", 2), Token.construct("three", 4)])
        ]),
        MatchableSegment([
            MatchableSegment([Token.construct("zero", 0), Token.construct("one", 1)]),
            MatchableSegment([Token.construct("two", 2), Token.construct("three", 3)])
        ]),
    )



"""
Apply operations
================
"""

from .operations import Insert, Delete, Equal


def apply(operations, a_tokens, b_tokens):
    """
    Applies a sequences of operations to tokens -- copies tokens from
    `a_tokens` and `b_tokens` according to `operations`.
    
    :Parameters:
        operations : sequence of :~class:`deltas.Operation`
            Operations to perform
        a_tokens : list of `comparable`
            Starting sequence of comparable tokens
        b_tokens : list of `comparable`
            Ending list of comparable tokens
        
    :Returns:
        A new list of tokens
    """
    for operation in operations:
        
        if isinstance(operation, Equal):
            #print("Equal: {0}".format(str(a_tokens[operation.a1:operation.a2])))
            for t in a_tokens[operation.a1:operation.a2]: yield t
        
        elif isinstance(operation, Insert):
            #print("Insert: {0}".format(str(b_tokens[operation.b1:operation.b2])))
            for t in b_tokens[operation.b1:operation.b2]: yield t
        
        elif isinstance(operation, Delete):
            #print("Delete: {0}".format(str(a_tokens[operation.a1:operation.a2])))
            pass
        
        else:
            raise TypeError("Unexpected operation type " + \
                            "{0}".format(type(operation)))



"""
Match segments
--------------

Performs a diffs using a tree of matchable segments in order to remain robust
to content moves.  This module supports the use of a custom
:class:`~deltas.segmenters.Segmenter`.

.. automethod:: deltas.algorithms.segment_matcher.diff

.. automethod:: deltas.algorithms.segment_matcher.diff_segments

.. automethod:: deltas.algorithms.segment_matcher.process

.. autoclass:: SegmentMatcher
    :members:
"""
from collections import defaultdict

from . import sequence_matcher
from ..operations import Delete, Equal, Insert
from ..segmenters import (MatchableSegment, ParagraphsSentencesAndWhitespace,
                          Segment, Segmenter)
from ..tokenizers import text_split, Token, Tokenizer
from .diff_engine import DiffEngine

SEGMENTER = ParagraphsSentencesAndWhitespace()
TOKENIZER = text_split

def diff(a, b, segmenter=None):
    """
    Performs a diff comparison between two sequences of tokens (`a` and `b`)
    using `segmenter` to cluster and match
    :class:`deltas.segmenters.MatchableSegment`.

    :Example:
        >>> from deltas import segment_matcher, text_split
        >>>
        >>> a = text_split.tokenize("This is some text.  This is some other text.")
        >>> b = text_split.tokenize("This is some other text.  This is some text.")
        >>> operations = segment_matcher.diff(a, b)
        >>>
        >>> for op in operations:
        ...     print(op.name, repr(''.join(a[op.a1:op.a2])), repr(''.join(b[op.b1:op.b2])))
        ...
        equal 'This is some other text.' 'This is some other text.'
        insert ' ' '  '
        equal 'This is some text.' 'This is some text.'
        delete '  ' ''

    :Parameters:
        a : `list`(:class:`deltas.tokenizers.Token`)
            Initial sequence
        b : `list`(:class:`deltas.tokenizers.Token`)
            Changed sequence
        segmenter : :class:`deltas.segmenters.Segmenter`
            A segmenter to use on the tokens.

    :Returns:
        An `iterable` of operations.
    """
    a, b = list(a), list(b)
    segmenter = segmenter or SEGMENTER

    # Cluster the input tokens
    a_segments = segmenter.segment(a)
    b_segments = segmenter.segment(b)

    return diff_segments(a_segments, b_segments)

def diff_segments(a_segments, b_segments):
    """
    Performs a diff comparison between two pre-clustered
    :class:`deltas.segmenters.Segment` trees.  In most cases, segmentation
    takes 100X more time than actually performing the diff.

    :Parameters:
        a_segments : :class:`deltas.segmenters.Segment`
            An initial sequence
        b_segments : :class:`deltas.segmenters.Segment`
            A changed sequence

    :Returns:
        An `iterable` of operations.
    """
    a_segments, b_segments = list(a_segments), list(b_segments)

    # Match and re-sequence unmatched tokens
    a_segment_tokens, b_segment_tokens = _cluster_matching_segments(a_segments,
                                                                    b_segments)

    # Perform a simple LCS over unmatched tokens and clusters
    clustered_ops = sequence_matcher.diff(a_segment_tokens, b_segment_tokens)

    # Return the expanded (de-clustered) operations
    return _expand_clustered_ops(clustered_ops,
                                 a_segment_tokens,
                                 b_segment_tokens)



def process(texts, *args, **kwargs):
    """
    Processes a single sequence of texts with a
    :class:`~diffengine.algoroithms.SegmentMatcher`.

    :Parameters:
        texts : `iterable`(`str`)
            sequence of texts
        args : `tuple`
            passed to :class:`~diffengine.algoroithms.SegmentMatcher`'s
            constructor
        kwaths : `dict`
            passed to :class:`~diffengine.algoroithms.SegmentMatcher`'s
            constructor
    """
    processor = SegmentMatcher.Processor(*args, **kwargs)
    for text in texts:
        yield processor.process(text)

class SegmentMatcher(DiffEngine):
    """
    Constructs a segment matcher diff engine that preserves segmentation state
    and is able to process changes sequentially.  When detecting changes
    across many versions of a text this strategy will be about twice as fast as
    calling :func:`diff` sequentially.

    :Example:
        >>> from deltas.algorithms import SegmentMatcher
        >>> from deltas.tokenizers import text_split
        >>>
        >>> engine = SegmentMatcher(text_split)
        >>>
        >>> processor = engine.processor()
        >>> ops, a, b = processor.process("This is a version.  It has some text in it.")
        >>> print(" ".join(repr(''.join(b[op.b1:op.b2])) for op in ops))
        'This is a version.  It has some text in it.'
        >>> ops, a, b = processor.process("This is a version.  However, it has different.")
        >>> print(" ".join(repr(''.join(b[op.b1:op.b2])) for op in ops))
        'This is a version.  ' '' 'However, it' ' has ' '' 'different' '.'
        >>> ops, a, b = processor.process("Switching it up here.  This is a version.")
        >>> print(" ".join(repr(''.join(b[op.b1:op.b2])) for op in ops))
        '' 'Switching' ' it ' '' 'up' ' ' '' 'here' '.' '  ' 'This is a version.'
    """

    class Processor(DiffEngine.Processor):
        """
        A processor used by the SegmentMatcher difference engine to track the
        history of a single text.
        """
        def __init__(self, tokenizer=None, segmenter=None, last_text=None,
                           last_tokens=None, last_segments=None):
            self.tokenizer = tokenizer or TOKENIZER
            self.segmenter = segmenter or SEGMENTER
            self.update(last_text, last_tokens, last_segments)

        def update(self, last_text=None, last_tokens=None, last_segments=None):
            if last_segments is not None:
                self.last_segments = last_segments
                self.last_tokens = self.last_segments.tokens()
            elif last_tokens is not None:
                self.last_tokens = last_tokens
                self.last_segments = self.segments.segment(last_tokens)
            elif last_text is not None:
                self.last_tokens = self.tokenizer.tokenize(last_text)
                self.last_segments = self.segmenter.segment(self.last_tokens)
            else:
                self.last_tokens = []
                self.last_segments = Segment()


        def process(self, text):
            """
            Processes a new version of a text and returns the delta.

            :Parameters:
                text : `str`
                    The text to process

            :Returns:
                    A tuple of `operations`, `a_tokens`, `b_tokens`
            """
            # Tokenize and segment
            tokens = self.tokenizer.tokenize(text)
            segments = self.segmenter.segment(tokens)

            return self.process_segments(segments, tokens=tokens)

        def process_segments(self, segments, tokens=None):

            if tokens is None: tokens = segments.tokens()

            # Perform diff
            operations = diff_segments(self.last_segments, segments)

            # Update state
            a = self.last_tokens
            b = tokens
            self.last_tokens = tokens
            self.last_segments = segments

            # Return delta
            return operations, a, b


    def __init__(self, tokenizer=None, segmenter=None):
        self.tokenizer = tokenizer or TOKENIZER
        self.segmenter = segmenter or SEGMENTER

    def processor(self, *args, **kwargs):
        """
        Constructs and configures a processor to process versions of a text.
        """
        return self.Processor(self.tokenizer, self.segmenter, *args, **kwargs)

    def process(self, texts, *args, **kwargs):
        return process(texts, self.tokenizer, self.segmenter, *args, **kwargs)

    @classmethod
    def from_config(cls, config, name, section_key="diff_engines"):
        section = config[section_key][name]
        return cls(
            Tokenizer.from_config(config, section['tokenizer']),
            Segmenter.from_config(config, section['segmenter'])
        )

def _cluster_matching_segments(a_segments, b_segments):

    # Generate a look-up map for matchable segments in 'a'
    a_segment_map = _build_segment_map(a_segments)

    # Find and cluster matching content in 'b'
    b_segment_tokens = list(_match_segments(a_segment_map, b_segments))

    # Expand unmatched segments from 'a'
    a_segment_tokens = list(_expand_unmatched_segments(a_segments))

    return a_segment_tokens, b_segment_tokens

def _expand_clustered_ops(operations, a_token_clusters, b_token_clusters):

    position = 0
    for operation in operations:
        if isinstance(operation, Equal):
            #print("Processing equal:")
            new_ops = _process_equal(position, operation,
                                     a_token_clusters, b_token_clusters)

        elif isinstance(operation, Insert):
            #print("Processing insert:")
            new_ops = _process_insert(position, operation,
                                      a_token_clusters, b_token_clusters)

        elif isinstance(operation, Delete):
            #print("Processing remove:")
            new_ops = _process_delete(position, operation,
                                      a_token_clusters, b_token_clusters)

        else:
            assert False, "Should never happen"

        for new_op in new_ops:

            yield new_op
            position = position + (new_op.b2 - new_op.b1)

def _build_segment_map(segments):
    d = defaultdict(list)
    for matchable_segment in _get_matchable_segments(segments):
        d[matchable_segment].append(matchable_segment)

    return d

def _get_matchable_segments(segments):
    """
    Performs a depth-first search of the segment tree to get all matchable
    segments.
    """
    for subsegment in segments:
        if isinstance(subsegment, Token):
            break # No tokens allowed next to segments
        if isinstance(subsegment, Segment):
            if isinstance(subsegment, MatchableSegment):
                yield subsegment

            for matchable_subsegment in _get_matchable_segments(subsegment):
                yield matchable_subsegment


def _match_segments(a_segment_map, b_segments):
    for subsegment in b_segments:
        if isinstance(subsegment, Segment):

            if isinstance(subsegment, MatchableSegment) and \
               subsegment in a_segment_map:
                matched_segments = a_segment_map[subsegment] # Get matches
                for matched_segment in matched_segments: # For each match
                    matched_segment.match = subsegment # flag as matched
                subsegment.match = matched_segments[0] # Always associate with first match
                yield subsegment # Dump matched segment

            else:
                for seg_or_tok in _match_segments(a_segment_map, subsegment):
                    yield seg_or_tok # Recurse

        else:
            yield subsegment # Dump token


def _expand_unmatched_segments(a_segments):
    for subsegment in a_segments:
        # Check if a segment is matched.
        if isinstance(subsegment, Segment):

            if isinstance(subsegment, MatchableSegment) and \
               subsegment.match is not None:
                yield subsegment # Yield matched segment as cluster
            else:
                for seg_or_tok in _expand_unmatched_segments(subsegment):
                    yield seg_or_tok # Recurse
        else:
            yield subsegment # Dump token

def _process_equal(position, operation, a_token_clusters, b_token_clusters):
    yield Equal(a_token_clusters[operation.a1].start,
                a_token_clusters[operation.a2-1].end,
                b_token_clusters[operation.b1].start,
                b_token_clusters[operation.b2-1].end)

def _process_insert(position, operation, a_token_clusters, b_token_clusters):
    inserted_tokens = []
    for token_or_segment in b_token_clusters[operation.b1:operation.b2]:

        if isinstance(token_or_segment, Token):
            inserted_tokens.append(token_or_segment)
        else: # Found a matched token.
            if len(inserted_tokens) > 0:
                yield Insert(inserted_tokens[0].start,
                             inserted_tokens[-1].end,
                             position,
                             position+len(inserted_tokens))

                # update & reset!
                position += len(inserted_tokens)
                inserted_tokens = []

            match = token_or_segment.match
            yield Equal(match.start, match.end,
                        position, position+(match.end-match.start))

            # update!
            position += match.end-match.start



    # cleanup
    if len(inserted_tokens) > 0:
        yield Insert(inserted_tokens[0].start,
                     inserted_tokens[-1].end,
                     position,
                     position+len(inserted_tokens))

def _process_delete(position, operation, a_token_clusters, b_token_clusters):
    removed_tokens = []
    for token_or_segment in a_token_clusters[operation.a1:operation.a2]:

        if isinstance(token_or_segment, Token):
            removed_tokens.append(token_or_segment)
        else: # Found a matched token... not removed -- just moved
            if len(removed_tokens) > 0:
                yield Delete(removed_tokens[0].start,
                             removed_tokens[-1].end,
                             position,
                             position)

            # update & reset!
            position += len(removed_tokens)
            removed_tokens = []

    # cleanup
    if len(removed_tokens) > 0:
        yield Delete(removed_tokens[0].start,
                     removed_tokens[-1].end,
                     position,
                     position)



"""
Tokens represent chuncks of text that have semantic meaning.  A Token class that
extends :class:`str` is profivided.

.. autoclass:: deltas.tokenizers.Token
    :members:
"""


class Token(str):

    __slots__ = ("type", "i", "start", "end")

    @classmethod
    def construct(cls, content, i, type=None):
        t = cls(content)
        t.i = i
        t.type = type
        t.start = i
        t.end = i+1
        return t

'''
    def __new__(cls, content, *args, **kwargs):
        if isinstance(content, cls):
            return content
        else:
            return super().__new__(cls, content)

    def __init__(self, content, i, type=None):
        super().__init__()
        self.type = str(type) if type is not None else None
        self.i = int(i)

    def __str__(self):
        return super().__str__()

    def __repr__(self):
        return "{0}({1}, i={2}, type={3})" \
               .format(self.__class__.__name__,
                       super().__repr__(),
                       self.i,
                       repr(self.type))

    @property
    def start(self):
        return self.i

    @property
    def end(self):
        return self.i+1
'''



"""
Segments represent subsequences of tokens that have interesting properties.  All
segments are based on two abstract types:

:class:`deltas.segmenters.segments.Segment`
    A segment of text with a ``start`` and ``end`` index that refers to the
    original sequence of tokens.
:class:`deltas.segmenters.segments.MatchableSegment`
    A segment of text that can be matched with another segment no matter where
    it appears in a document.  Generally segmnents of this type represent a
    substantial collection of tokens.

Segment Types
^^^^^^^^^^^^^

.. autoclass:: deltas.segmenters.segments.Segment
    :members:

.. autoclass:: deltas.segmenters.segments.MatchableSegment
    :members:
"""
import hashlib
import types

from ..util import LookAhead


class Segment(list):
    """
    Represents a sequence of of tokens.  Note that plain Segments are not
    matchable.  For matchable segments, see
    :class:`deltas.segmenters.MatchableSegment`.
    """
    def __new__(cls, *args, **kwargs):
        if len(args) == 1 and len(kwargs) == 0 and isinstance(args[0], cls):
            args[0]
        else:
            inst = super().__new__(cls, *args, **kwargs)
            inst.initialize(*args, **kwargs)
            return inst

    def __init__(self, *args, **kwargs): pass

    def initialize(self, subsegments=None):
        subsegments = subsegments or []
        super().__init__(subsegments)

    def tokens(self):
        for subsegment_or_token in self:
            if isinstance(subsegment_or_token, Segment):
                subsegment = subsegment_or_token
                for token in subsegment.tokens():
                    yield token
            else:
                token = subsegment_or_token
                yield token

    @property
    def start(self):
        """
        The :func:`deltas.tokenizers.Token.start` of the first
        :class:`deltas.tokenizers.Token` in the segment.
        """
        return self[0].start

    @property
    def end(self):
        """
        The :func:`deltas.tokenizers.Token.end` of the last
        :class:`deltas.tokenizers.Token` in the segment.
        """
        return self[-1].end

    def __repr__(self):
        return "{0}({1})".format(self.__class__.__name__, super().__repr__())

    def __str__(self):
        return ''.join(str(ss) for ss in self)

    def __eq__(self, other):
        raise NotImplementedError()

    def __neq__(self, other):
        raise NotImplementedError()

    def __hash__(self, other):
        raise NotImplementedError()


class MatchableSegment(Segment):
    __slots__ = ("sha1", "match")

    def initialize(self, subsegments=None):
        super().initialize(subsegments)
        self.sha1 = hashlib.sha1(bytes(str(self), 'utf-8'))
        self.match = None

    def __eq__(self, other):
        try:
            return hash(self) == hash(other)
        except AttributeError:
            return False

    def __neq__(self, other):
        try:
            return hash(self) != hash(other)
        except AttributeError:
            return False

    def __hash__(self):
        return hash(self.sha1.digest())

    def append(self, subsegment):
        super().append(subsegment)
        self.sha1.update(bytes(str(subsegment), 'utf-8'))

    def extend(self, subsegments):
        for subsegment in subsegments:
            self.append(subsegment)
